@@ -1,7 +1,11 @@
-from ._beta_bernoulli_mixture import BetaBernoulliMixture
-from ._beta_binomial_mixture import BetaBinomialMixture
-from ._beta_in_beta_mixture import BetaInBetaMixture
+from ._base import BaseGaussianMixture
 from ._dirichlet_distribution_mixture import DirichletDistributionMixture
 from ._dirichlet_process_mixture import DirichletProcessMixture
-from ._geometric_process_mixture import GeometricProcessMixture
 from ._pitman_yor_process_mixture import PitmanYorMixture
+from ._geometric_process_mixture import GeometricProcessMixture
+from ._beta_in_beta_mixture import BetaInBetaMixture
+from ._beta_in_dirichlet_mixture import BetaInDirichletMixture
+from ._beta_bernoulli_mixture import BetaBernoulliMixture
+from ._beta_binomial_mixture import BetaBinomialMixture
+from ._equal_weighted_mixture import EqualWeightedMixture
+from ._frequency_weighted_mixture import FrequencyWeightedMixture
