@@ -1,5 +1,6 @@
 from sklearn.cluster import SpectralClustering
 from scipy.stats import multivariate_normal
+from collections import defaultdict
 import pandas as pd
 import numpy as np
 
@@ -9,6 +10,7 @@ from . import _utils
 from ..exceptions import NotFittedError
 from ..weight_models import BaseWeight
 from ..utils.functions import density_students_t, density_normal
+from ..utils.validators import rng_parser
 
 
 class BaseGaussianMixture(metaclass=ABCMeta):
@@ -53,15 +55,10 @@ class BaseGaussianMixture(metaclass=ABCMeta):
                  lambda_prior=1, psi_prior=None, nu_prior=None,
                  total_iter=1000, burn_in=100, subsample_steps=1,
                  show_progress=False, rng=None):
-        if rng is None:
-            self.rng = np.random.default_rng()
-        elif type(rng) is int:
-            self.rng = np.random.default_rng(rng)
-        else:
-            self.rng = rng
 
         assert total_iter > burn_in, (
             "total_iter must be greater than burn_in period")
+        self.rng = rng_parser(rng)
         self.burn_in = int(burn_in)
         self.total_iter = int(total_iter)
         self.subsample_steps = int(subsample_steps)
@@ -76,8 +73,7 @@ class BaseGaussianMixture(metaclass=ABCMeta):
 
         # Variables used in Gibbs sampler
         self.d = np.array([])
-        self.mu = np.array([])
-        self.sigma = np.array([[]])
+        self.theta = {}
         self.u = np.array([])
         self.affinity_matrix = np.array([])
         self.map_sim_params = None
@@ -88,7 +84,7 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         self.n_atoms = []
         self.n_log_likelihood = []
 
-        # Variables used in variational method
+        # Extra variables used in variational methods
         self.var_k = None
         self.var_d = None
         self.var_theta = None
@@ -96,6 +92,7 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         # Fitting flags
         self.gibbs_fitted = False
         self.var_fitted = False
+        self.var_converged = False
 
         self.show_progress = show_progress
 
@@ -135,28 +132,31 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         self._initialize_common_params(y)
         if not warm_start:
             self._initialize_gibbs_params(max_groups=max_groups, method=method)
+            self._update_map_params()
         if show_progress is not None:
             self.show_progress = show_progress
+
         # Iterate the Gibbs steps with or without tqdm
+        burn_in_iterator = range(self.burn_in)
+        range_iterator = range(self.total_iter - self.burn_in)
         if self.show_progress:
+            from tqdm import tqdm
+            burn_in_iterator = tqdm(burn_in_iterator)
+            range_iterator = tqdm(range_iterator)
             print("Starting burn-in.")
-            from tqdm import trange
-            for _ in trange(self.burn_in):
-                self._gibbs_step()
+        for _ in burn_in_iterator:
+            self._gibbs_step()
+            self._update_map_params()
+        if self.show_progress:
             print("Finished burn-in.")
             print("Starting training.")
-            for i in trange(self.total_iter - self.burn_in):
-                self._gibbs_step()
-                if i % self.subsample_steps == 0:
-                    self._save_params()
+        for i in range_iterator:
+            self._gibbs_step()
+            self._update_map_params()
+            if i % self.subsample_steps == 0:
+                self._save_params()
+        if self.show_progress:
             print("Finished training.")
-        else:
-            for _ in range(self.burn_in):
-                self._gibbs_step()
-            for i in range(self.total_iter - self.burn_in):
-                self._gibbs_step()
-                if i % self.subsample_steps == 0:
-                    self._save_params()
         self.gibbs_fitted = True
 
     def fit_variational(self, y, n_groups=None, warm_start=False,
@@ -205,35 +205,31 @@ class BaseGaussianMixture(metaclass=ABCMeta):
             self.show_progress = show_progress
         self._initialize_common_params(y)
         if not warm_start:
-            if n_groups is None:
-                if hasattr(self, 'n'):
-                    var_k = self.n
-                else:
-                    raise AttributeError("n_groups must be a positive integer")
+            if hasattr(self, 'n') and n_groups is None:
+                var_k = self.n
+            elif n_groups is None:
+                raise AttributeError("n_groups must be a positive integer")
             else:
                 var_k = n_groups
             self._initialize_variational_params(var_k=var_k, method=method)
         elbo = -np.inf
         elbo_diff = np.inf
         iterations = 0
+        t = None
         if self.show_progress:
             from tqdm import tqdm
-            with tqdm() as t:
-                while elbo_diff > tol and iterations < max_iter:
-                    self._maximize_variational()
-                    prev_elbo = elbo
-                    elbo = self._calc_elbo()
-                    elbo_diff = abs(prev_elbo - elbo)
-                    iterations += 1
-                    t.update()
-        else:
-            while elbo_diff > tol and iterations < max_iter:
-                self._maximize_variational()
-                prev_elbo = elbo
-                elbo = self._calc_elbo()
-                elbo_diff = abs(prev_elbo - elbo)
-                iterations += 1
+            t = tqdm()
+        while elbo_diff > tol and iterations < max_iter:
+            self._maximize_variational()
+            prev_elbo = elbo
+            elbo = self._calc_elbo()
+            elbo_diff = abs(prev_elbo - elbo)
+            iterations += 1
+            if t is not None:
+                t.update()
         self.var_fitted = True
+        if iterations < max_iter:
+            self.var_converged = True
 
     def gibbs_eap_density(self, y=None, periods=None):
         """
@@ -258,20 +254,15 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         _y = self._cast_observations(y)
         y_sim = []
         if periods is None:
-            for param in self.sim_params:
-                y_sim.append(_utils.mixture_density(_y,
-                                                    param["w"],
-                                                    param["mu"],
-                                                    param["sigma"],
-                                                    param["u"]))
+            sim_params = self.sim_params
         else:
-            periods = min(periods, len(self.sim_params))
-            for param in self.sim_params[-periods:]:
-                y_sim.append(_utils.mixture_density(_y,
-                                                    param["w"],
-                                                    param["mu"],
-                                                    param["sigma"],
-                                                    param["u"]))
+            i_start = min(periods, len(self.sim_params))
+            sim_params = self.sim_params[-i_start:]
+        for param in sim_params:
+            y_sim.append(_utils.mixture_density(_y,
+                                                param["w"],
+                                                param["theta"],
+                                                param["u"]))
         return np.array(y_sim).mean(axis=0)
 
     def gibbs_map_density(self, y=None):
@@ -293,8 +284,7 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         _y = self._cast_observations(y)
         return _utils.mixture_density(_y,
                                       self.map_sim_params["w"],
-                                      self.map_sim_params["mu"],
-                                      self.map_sim_params["sigma"],
+                                      self.map_sim_params["theta"],
                                       self.map_sim_params["u"])
 
     def gibbs_eap_affinity_matrix(self, y=None):
@@ -321,8 +311,7 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         for params in self.sim_params:
             grouping = _utils.cluster(_y,
                                       params["w"],
-                                      params["mu"],
-                                      params["sigma"],
+                                      params["theta"],
                                       params["u"])[0]
             affinity_matrix += np.equal(grouping, grouping[:, None])
         affinity_matrix /= len(self.sim_params)
@@ -373,8 +362,7 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         _y = self._cast_observations(y)
         ret = _utils.cluster(_y,
                              self.map_sim_params["w"],
-                             self.map_sim_params["mu"],
-                             self.map_sim_params["sigma"],
+                             self.map_sim_params["theta"],
                              self.map_sim_params["u"])
         if not full:
             ret = ret[0]
@@ -454,6 +442,10 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         y : {array-like} of shape (n_samples, n_features), default=None
             The points to cluster using the MAP assignations. If `None`
             the data used at fitting is used.
+        full: bool, default=False
+            If False (default), only the maximum a posteriori clustering is
+            returned. If True, the variational assignation probabilty is also
+            returned.
         """
         if not self.var_fitted:
             raise NotFittedError("Object must be fitted with fit_variational"
@@ -533,18 +525,29 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         max_groups: int, default=None
             Maximum number of groups to assign in the initialization. If None,
             the  number of groups drawn from the weight model is not caped.
-        method: str, default="random"
+        method: str
             "kmeans": does a kmeans initialization
             "random": does a random initialization based on the prior models
             "variational": fits the variational distribution an uses the MAP
                 parameters as initialization
         """
-        self.mu = np.empty((0, *self.mu_prior.shape))
-        self.sigma = np.empty((0, *self.psi_prior.shape))
+
+        def atom_generator():
+            mu, sigma = _utils.random_normal_invw(
+                mu=self.mu_prior,
+                lam=self.lambda_prior,
+                psi=self.psi_prior,
+                nu=self.nu_prior,
+                rng=self.rng
+            )
+            return np.atleast_1d(mu), np.atleast_2d(sigma)
+
         self.sim_params = []
         self.n_groups = []
         self.n_atoms = []
         self.total_saved_steps = 0
+
+        self.theta = defaultdict(atom_generator)
         self.affinity_matrix = np.zeros((len(self.y), len(self.y)))
         self.u = self.rng.uniform(0 + np.finfo(np.float64).eps, 1,
                                   len(self.y))
@@ -554,14 +557,11 @@ class BaseGaussianMixture(metaclass=ABCMeta):
             self.weight_model.complete(max_groups)
 
         if method == "kmeans":
-            from sklearn.cluster import KMeans
-            # TODO wait for sklearn to implement Generator as random_state
-            # input https://github.com/scikit-learn/scikit-learn/issues/16988
-            km = KMeans(
-                n_clusters=self.var_k,
-                random_state=np.random.RandomState(self.rng.bit_generator)
-            )
-            self.d = km.fit_predict(self.y)
+            if hasattr(self, 'n'):
+                n = self.n
+            else:
+                n = 10
+            self.d = _utils.kmeans_cluster_size_biased(self.y, n, self.rng)
         elif method == "random":
             self.d = self.weight_model.random_assignment(len(self.y))
         elif method == "variational":
@@ -571,7 +571,6 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         else:
             raise AttributeError("method param must be one of 'kmeans', "
                                  "'random', 'variational'")
-        self._complete_atoms()
 
     def _initialize_variational_params(self, var_k, method="kmeans"):
         """
@@ -596,45 +595,15 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         self.var_theta = []
 
         if method == "kmeans":
-            from sklearn.cluster import KMeans
-            # TODO wait for sklearn to implement Generator as random_state
-            # input https://github.com/scikit-learn/scikit-learn/issues/16988
-            km = KMeans(
-                n_clusters=self.var_k,
-                random_state=np.random.RandomState(self.rng.bit_generator)
-            )
-            d = km.fit_predict(self.y)
-            dim = self.y.shape[1]
+            d = _utils.kmeans_cluster_size_biased(self.y, self.var_k,
+                                                  self.rng)
             var_d = np.zeros((self.var_k, self.y.shape[0]),
                              dtype=np.float64)
-            for j in range(self.var_k):
-                y_subset = self.y[d == j]
-                if len(y_subset) == 0:
-                    var_d[j, :] = np.finfo(np.float64).eps
-                    continue
-                mu_j = np.mean(y_subset, 0)
-                precisions_j = np.linalg.inv(
-                    np.atleast_2d(np.cov(y_subset.T))
-                )
-                mu_j = np.atleast_1d(mu_j)
-                precisions_j = np.atleast_2d(precisions_j)
-                # atoms initialization
-                self.var_theta.append([mu_j, self.lambda_prior,
-                                       precisions_j, self.nu_prior])
-                # assignations initialization
-                log_d_ji = _utils.e_log_norm_wishart(precisions_j,
-                                                     self.nu_prior) / 2
-                log_d_ji -= dim / (2 * self.lambda_prior)
-                log_d_ji -= (self.nu_prior / 2 *
-                             ((self.y - mu_j).T * (precisions_j @ (
-                                     self.y - mu_j).T)).sum(0)
-                             )
-                var_d[j, :] = log_d_ji
-            var_d -= var_d.max(0)
-            var_d = np.exp(var_d)
-            var_d += np.finfo(np.float64).eps
-            var_d /= var_d.sum(0)
+            var_d[d, range(len(d))] = 1
             self.var_d = var_d
+            self.weight_model.fit_variational(np.empty(shape=(self.var_k, 0)))
+            self._update_var_theta()
+            self._update_var_d()
         elif method == "random":
             for _ in range(self.var_k):
                 mu_j, temp_psi = _utils.random_normal_invw(
@@ -656,26 +625,27 @@ class BaseGaussianMixture(metaclass=ABCMeta):
 
     def _get_run_params(self):
         return {"w": self.weight_model.get_weights(),
-                "mu": self.mu,
-                "sigma": self.sigma,
+                "theta": dict(self.theta),
                 "u": self.u,
                 "d": self.d}
 
     def _save_params(self):
         self.sim_params.append(self._get_run_params())
         self.n_groups.append(len(np.unique(self.d)))
-        self.n_atoms.append(len(self.mu))
-        self._update_map_params()
+        self.n_atoms.append(len(self.theta))
         self.affinity_matrix += np.equal(self.d, self.d[:, None])
         self.total_saved_steps += 1
 
     def _update_map_params(self):
-        run_log_likelihood = self._y_log_likelihood()
+        """Calc the likelihood and parameters of the run. Update MAP if the
+        likelihood is greater
+        """
+        run_log_likelihood = self._run_log_likelihood()
         if self.map_log_likelihood < run_log_likelihood:
             self.map_log_likelihood = run_log_likelihood
             self.map_sim_params = self._get_run_params()
         elif self.map_log_likelihood == -np.inf:
-            # It's better than nothing
+            # Save the params to get something to compare
             self.map_sim_params = self._get_run_params()
 
     def _update_weights(self):
@@ -686,15 +656,10 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         self.weight_model.tail(1 - min(self.u))
 
     def _update_atoms(self):
-        assert len(self.mu) == len(self.sigma)
-        self.d = np.unique(self.d, return_inverse=True)[1]
-        self.mu = []
-        self.sigma = []
-        for j in range(max(self.d) + 1):
-            inj = (self.d == j).nonzero()[0]
-
+        for j in np.unique(self.d):
+            mask_j = self.d == j
             posterior_params = _utils.posterior_norm_invw_params(
-                self.y[inj],
+                self.y[mask_j],
                 mu=self.mu_prior,
                 lam=self.lambda_prior,
                 psi=self.psi_prior,
@@ -705,25 +670,9 @@ class BaseGaussianMixture(metaclass=ABCMeta):
                 psi=posterior_params["psi"],
                 nu=posterior_params["nu"],
                 rng=self.rng)
-            self.mu.append(temp_mu)
-            self.sigma.append(temp_sigma)
-
-        self.mu = np.array(self.mu).reshape(-1, *self.mu_prior.shape)
-        self.sigma = np.array(self.sigma).reshape(-1, *self.psi_prior.shape)
-
-    def _complete_atoms(self):
-        missing_len = self.weight_model.get_size() - len(self.mu)
-        for _ in range(missing_len):
-            temp_mu, temp_sigma = _utils.random_normal_invw(
-                mu=self.mu_prior,
-                lam=self.lambda_prior,
-                psi=self.psi_prior,
-                nu=self.nu_prior,
-                rng=self.rng
-            )
-            self.mu = np.concatenate((self.mu, [np.atleast_1d(temp_mu)]))
-            self.sigma = np.concatenate((self.sigma,
-                                         [np.atleast_2d(temp_sigma)]))
+            temp_mu = np.atleast_1d(temp_mu)
+            temp_sigma = np.atleast_2d(temp_sigma)
+            self.theta[j] = (temp_mu, temp_sigma)
 
     def _update_d(self):
         log_prob = self._d_log_likelihood_vector()
@@ -732,15 +681,14 @@ class BaseGaussianMixture(metaclass=ABCMeta):
     def _gibbs_step(self):
         self._update_atoms()
         self._update_weights()
-        self._complete_atoms()
         self._update_d()
 
     def _d_log_likelihood_vector(self):
         with np.errstate(divide='ignore'):
             log_probability = np.array(
                 [multivariate_normal.logpdf(self.y,
-                                            self.mu[j],
-                                            self.sigma[j],
+                                            self.theta[j][0],
+                                            self.theta[j][1],
                                             1)
                  for j in range(self.weight_model.get_size())]
             )
@@ -749,16 +697,48 @@ class BaseGaussianMixture(metaclass=ABCMeta):
                 self.u))
         return log_probability
 
+    def _run_log_likelihood(self):
+        ret = 0
+        ret += self._y_log_likelihood()
+        ret += self._d_log_likelihood()
+        ret += self._w_log_likelihood()
+        ret += self._theta_log_likelihood()
+        return ret
+
     def _y_log_likelihood(self):
+        """returns the loglikelihood of f(y|d, w, theta)"""
         ret = 0
         with np.errstate(divide='ignore'):
-            for di in np.unique(self.d):
-                ret += np.sum(multivariate_normal.logpdf(self.y[self.d == di],
-                                                         self.mu[di],
-                                                         self.sigma[di],
+            for j in np.unique(self.d):
+                ret += np.sum(multivariate_normal.logpdf(self.y[self.d == j],
+                                                         self.theta[j][0],
+                                                         self.theta[j][1],
                                                          1))
         return ret
 
+    def _d_log_likelihood(self):
+        """returns the loglikelihood of f(d|w)"""
+        return self.weight_model.assignation_log_likelihood(self.d)
+
+    def _w_log_likelihood(self):
+        """returns the loglikelihood of f(w)"""
+        return self.weight_model.weighting_log_likelihood()
+
+    def _theta_log_likelihood(self):
+        """returns the loglikelihood of f(theta)"""
+        res = 0
+        for j in np.unique(self.d):
+            mu, sigma = self.theta[j]
+            res += _utils.log_likelihood_normal_invw(
+                mu=mu,
+                sigma=sigma,
+                mu0=self.mu_prior,
+                lam0=self.lambda_prior,
+                psi0=self.psi_prior,
+                nu0=self.nu_prior
+            )
+        return res
+
     def _maximize_variational(self):
         self._update_var_d()
         self._update_var_w()
@@ -783,7 +763,9 @@ class BaseGaussianMixture(metaclass=ABCMeta):
         for vd_j in self.var_d:
             n_j = vd_j.sum()
             x_bar_j = (vd_j / n_j) @ self.y
-            ns_j = (vd_j * (self.y - x_bar_j).T) @ (self.y - x_bar_j)
+            ns_j = np.einsum('i,ij,ik->jk',
+                             vd_j, (self.y - x_bar_j), (self.y - x_bar_j)
+                             ) / n_j
             v_lambda_j = self.lambda_prior + n_j
             v_scale_j = self.nu_prior + n_j
             v_mu_j = (self.lambda_prior * self.mu_prior +
@@ -804,16 +786,17 @@ class BaseGaussianMixture(metaclass=ABCMeta):
             v_mu_j, v_lambda_j, v_precision_j, v_scale_j = vt_j
             log_d_ji = self.weight_model.variational_mean_log_w_j(j)
             log_d_ji += _utils.e_log_norm_wishart(v_precision_j, v_scale_j) / 2
-            log_d_ji -= dim / (2 * v_lambda_j)
-            log_d_ji -= (v_scale_j / 2 *
+            log_d_ji -= dim * np.log(2 * np.pi) / 2
+            log_d_ji -= dim / v_lambda_j / 2
+            log_d_ji -= (v_scale_j *
                          ((self.y - v_mu_j).T * (
                                  v_precision_j @ (self.y - v_mu_j).T)).sum(0)
-                         )
+                         ) / 2
             var_d[j, :] = log_d_ji
-        var_d -= var_d.max(0)
+        var_d -= var_d.max(axis=0, initial=-np.inf)
         var_d = np.exp(var_d)
         var_d += np.finfo(np.float64).eps
-        var_d /= var_d.sum(0)
+        var_d /= var_d.sum(axis=0)
         self.var_d = var_d
 
     def _e_q_log_p_x(self):
@@ -901,4 +884,4 @@ class BaseGaussianMixture(metaclass=ABCMeta):
                 _y = np.array([[y]])
             else:
                 raise TypeError("Invalid type for variable y")
-        return _y
\ No newline at end of file
+        return _y
